#include "stdafx.h"
#include <cstdio>
#include <string>
#include <sstream>
#include <iostream>
#include <iterator>
#include <fstream>
using namespace std;
#include "ImageFileFormat.h"
#include "Image.h"

#include "DataStructures.h"
#include "Controller.h"
extern DebugLogFile g_debug_logfile;

// !!! Note:  The frequent use of CFileFind in this class may be slowing down the program unnecessarily.  
// Efficiency could be improved by loading all image file names into a vector, then accessing the vector
// rather than searching the disk each time. -- STB 4/27/10

// !!! The aforementioned filename storage should hash filenames rather than just push onto an unordered vector? -- ADP

ImageProgressBar::ImageProgressBar(const CString &text, int num_files, CWnd *pParentWnd)
: m_dc_ptr(NULL), m_lower(-1), m_upper(-1)
{
  if(!pParentWnd)return;
  CClientDC dc(pParentWnd);
  m_dc_ptr = &dc;
  m_dc_ptr->SetBkColor(::GetSysColor(COLOR_3DFACE));
  pParentWnd->GetWindowRect(m_rect[0]);
  pParentWnd->ScreenToClient(m_rect[0]);
  m_rect[1] = CRect(m_rect[0].left+(m_rect[0].Width()/2)-150, m_rect[0].top+(m_rect[0].Height()/2)-15, m_rect[0].left+(m_rect[0].Width()/2)+150, m_rect[0].top+(m_rect[0].Height()/2)+15);
  m_rect[2] = CRect(m_rect[1].left,m_rect[1].top-20,m_rect[1].right,m_rect[1].bottom-20);
  m_dc_ptr->DrawText(text,m_rect[2],0);
  m_progress_bar.Create(WS_CHILD | WS_VISIBLE | PBS_SMOOTH , m_rect[1], pParentWnd,0x16);
  m_lower = 0;
  m_upper = num_files;
  m_progress_bar.SetRange(m_lower,m_upper);
  m_progress_bar.SetStep(1);
}

ImageProgressBar::~ImageProgressBar()
{
  if(!m_dc_ptr)return;
  CBrush background_brush(::GetSysColor(COLOR_3DFACE));
  m_dc_ptr->FillRect(m_rect[2],&background_brush);
  m_progress_bar.ShowWindow(SW_HIDE);
}

void ImageProgressBar::StepIt(bool wrap_around)
{
  if(wrap_around)
  {
    if(m_progress_bar.GetPos() == m_upper)
      m_progress_bar.SetPos(m_lower);
  }
  m_progress_bar.StepIt();
}

/* returns a CString representation of an int */
CString itos(int i)
{
  stringstream ss;
  ss << i;
  return ss.str().c_str();
}

/* returns a CString representation of the tube, window, and session for debugging purposes */
CString GetTWSString(const ImageInfo& ii)
{
  CString s;
  s.Format("Tube: %s\nWindow: %d\nSession: %d",ii.tube,ii.window,ii.session);
  return s;
}

/* Constructor - initializes dialog to Bartz format */
ImageFileFormat::ImageFileFormat()
{
  SetToBartz();
}

bool ImageFileFormat::operator==(const ImageFileFormat& other) const
{
  if(! (m_image_file_format_info == other.m_image_file_format_info) ) return false;
  
  return true;
  
}

void ImageFileFormat::SetFileDirectory(const CString& directory)
{
  SetLengths(directory);
  m_map.clear();
}

CString ImageFileFormat::GetFileName(const CString& dir, const CString& tube, int window, int session) const
{ 
  // if valid inputs
  if(tube.GetLength() > 0 && window > 0 && session > 0)
  {
    
    CString file_format = "";  ///< used to search for the file
    CString file_name = "";    ///< the filename to return
    CString tmp_str = "";      ///< a temp string object
    OrderElement tmp;          ///< a temp order element
    
    // create file format string
    for(int i = 0; i < NUM_OF_ELEMENTS; i++)
    {
      tmp = m_image_file_format_info.InfoOrder[i];
      if(tmp.included)
      {
        // add prefix first
        file_format += tmp.prefix;
        
        // add elements depending on type
        if(tmp.element == WINDOW)
        {
          tmp_str.Format("%0" + itos(tmp.length) + "d",window);
          file_format += tmp_str;
        }
        else if(tmp.element == SESSION)        
        {
          tmp_str.Format("%0" + itos(tmp.length) + "d",session);
          file_format += tmp_str;
        }
        else if(tmp.element == TUBE)
        {
          tmp_str.Format("%s",tube);
          file_format += tmp_str;
        }
        else // use wildcard (*) for other elements
        {
          file_format += "*";
        }
        
        // add element delimiter
        file_format += m_image_file_format_info.delim;
      } // if included
    } // for number of elements in filename
    
    // replace last delimiter with a period, then add proper file extension
    file_format.SetAt(file_format.GetLength()-1,'.');
    file_format += m_image_file_format_info.filetype;
    
    CString searchstr;           ///< search string
    
    // add path information
    searchstr = dir+file_format;
    
    // search for files
    CFileFind ff;              ///< used to find files
    if(ff.FindFile(searchstr))
    {
      ff.FindNextFile();
      file_name = dir+ff.GetFileName();
    } // if file found
    return file_name;
  }
  
  return "";
}


CString ImageFileFormat::GetImageFileName(const CString& directory, const CString& tube, int window, int session) const
{
  CString file_name = "";    // the filename to return

  // if valid inputs
  if(tube.GetLength() > 0 && window > 0 && session > 0)
  {
    // build the key struct
    Key image_key(tube, window, session);

    // follow the map to the file name
    ImageFileMap::const_iterator it = m_map.find(image_key);
    if (it != m_map.end())
    {
      file_name = directory + (it->second); 
    }
    else
    {
      // Else block could possibly be removed.
      // There should not be any filenames that this function cannot 
      // find and the old one can.
      // It is here because C_CreateNewProject gets called before 
      // SetNavigation in OnFileNew. Because of this, LoadAll is called
      // before the map is built in CatalogueImages.
      // If we take it out now, there are some issues with asserts 
      // failing because rootfly thinks an image is loaded in the buffer
      // and it really is not. -CDA 04/29/2010
      file_name = GetFileName(directory,tube,window,session);
    }
  }

  return file_name;
}

CString ImageFileFormat::GetFileSearchString() const
{
  CString file_format = "";
  OrderElement tmp;
  
  for(int i = 0; i < NUM_OF_ELEMENTS; i++)
  {
    tmp = m_image_file_format_info.InfoOrder[i];
    if(tmp.included)
    {
      file_format += tmp.prefix;
      file_format += "*";
      file_format += m_image_file_format_info.delim;
    } // if included
  } // for number of elements in filename
  
  file_format.SetAt(file_format.GetLength()-1,'.');
  file_format += m_image_file_format_info.filetype;
  
  return file_format;
}

// Given a filename preceded by a path, removes the path.  Returns just the filename.
CString StripPath(const CString& filename_with_path)
{
  // Remove path to file if it exists
  int tmpI1 = filename_with_path.ReverseFind('\\');
  int tmpI2 = filename_with_path.ReverseFind('/');
  
  // If path information exists
  if (!(tmpI1 == -1 && tmpI2 == -1))
  {
    // Set tmpI1 to location of last 'slash'
    if (tmpI1 < tmpI2) tmpI1 = tmpI2;
    
    // remove path information
    return filename_with_path.Right(filename_with_path.GetLength()-(tmpI1+1));
  }
  else
  {
    return filename_with_path;
  }
}


bool ImageFileFormat::SetLengths(const CString& directory)
{
  CString file_name = GetFirstImage( directory );   // file name of first valid file found
  if (file_name == "")  return false;  // no file found
  
  OrderElement tmpOE;     ///< temp order element
  CString tmpStr;         ///< temp string
  int plen, delim, elen;  ///< length of prefix, location of delim, length of element
  
  // remove path, leaving only filename
  file_name = StripPath( file_name );
  
  try 
  {
    for (int i = 0; i < NUM_OF_ELEMENTS; i++)
    {
      tmpOE = m_image_file_format_info.InfoOrder[i];
      if(tmpOE.included)
      {
        // find next delimiter
        delim = file_name.Find(m_image_file_format_info.delim);
        
        // set cut to end if last element
        if(delim == -1)
        { 
          if (i == m_image_file_format_info.numElementsInFileName - 1)  delim = file_name.GetLength();
          else
          {
            assert(0);  // ran out of elements too soon.  Bad image file format.
            return false;
          }
        }
        
        // get length of prefix
        plen = tmpOE.prefix.GetLength();
        
        // extract element minus prefix, set element length
        elen = delim - plen;
        tmpStr = file_name.Mid(plen,elen);
        
        // set lengths here!
        if(tmpStr.Right(4) == ".jpg") // 4 is length of ".jpg"
          m_image_file_format_info.InfoOrder[i].length = tmpStr.GetLength() - 4;
        else
          m_image_file_format_info.InfoOrder[i].length = tmpStr.GetLength();
        
        // cut off previous element
        file_name = file_name.Right(file_name.GetLength()-(delim+1));
      } // if element is included
    } // for number of elements in file
  }// try this
  catch (...) 
  {
    TRACE("Error reading image file.\n\nThe file that caused the error was:\n'"+file_name+"'");
    assert(0);  // this should never happen
    return false;
  }
  return true;
}

bool ImageFileFormat::ImageFileExistsInDirectory(const CString& directory) const
{
  CFileFind ff;
  
  return ff.FindFile(directory + GetFileSearchString()) ? true : false;
}


int ImageFileFormat::CountFilesInDirectory(const CString& directory) const
{
  int count = 0;
  
  CFileFind ff;
  int more = ff.FindFile(directory + GetFileSearchString());
  
  while (more)
  {
    more = ff.FindNextFile();
    // skip over the dot directories . and ..
    if (!ff.IsDots() && !ff.IsDirectory())  count++;
  }
  
  return count;
}
  

// iterates over all files in the directory, noting navigation limits and existing files while updating a progress bar
// !!! This function directly manipulates the main window - it probably no longer belongs in this file
bool ImageFileFormat::CatalogueImages(const CString& directory, int* max_window, int* min_window,
                                      int* max_session,int* min_session, CWnd *pParentWnd, Controller &controller) const
{  // TODO: this function directly manipulates the main window - they probably no longer belong in this file
  CFileFind ff;
  ImageInfo ii;
  
  // progress bar manipulation
  ImageProgressBar ipb("Finding Images In Directory...         ", CountFilesInDirectory(directory), pParentWnd);

  // Since m_map doesn't really belong in this class anyway (need to refactor), we cast away constness
  ImageFileMap* map = const_cast<ImageFileMap*>( &m_map );

  // clear the map so we can build a new one
  map->clear();
  
  int more = ff.FindFile(directory + GetFileSearchString());
  if(!more) return false;
  
  while (more)
  {
    more = ff.FindNextFile();
    ipb.StepIt(true);
    ii = GetImageInfo(ff.GetFileName());
    if(ii.window > *max_window) *max_window = ii.window;
    if(ii.window < *min_window) *min_window = ii.window;
    if(ii.session > *max_session) *max_session = ii.session;
    if(ii.session < *min_session) *min_session = ii.session;
    
    controller.C_SetImageExists(true, ii.tube, ii.window, ii.session);

    // insert the filename into the map
    Key image_key(ii.tube, ii.window, ii.session);
    map->insert(make_pair(image_key, ff.GetFileName()));
  }
  return true;
}

// compose search string function
vector<Date> ImageFileFormat::GetDatesFromDir(const CString& dir, int max_sessions, CWnd *pParentWnd) const
{
  CString temp_str = "", temp;
  OrderElement tmp;
  CFileFind ff;
  ImageInfo ii;
  Date temp_date;
  vector<Date> dates_from_file;
  
  ImageProgressBar ipb("Reading Dates From File...         ", max_sessions, pParentWnd);
  
  // finds all dates in file
  for(int index = 1; index <= max_sessions; index++)
  {
    ipb.StepIt(false);
    // gets the search string
    temp_str = "";
    for(int i = 0; i < NUM_OF_ELEMENTS; i++)
    {
      tmp = m_image_file_format_info.InfoOrder[i];  
      if(tmp.included)
      {
        if(tmp.element == SESSION)
        {
          CString foo;
          foo.Format("%%0%dd", tmp.length);  // creates "%03d", for example -- STB 11/13/09
          temp.Format(foo,index);            // creates "001" for session 1, etc. -- STB 11/13/09
          temp_str += tmp.prefix;
          temp_str += temp;
          temp_str += m_image_file_format_info.delim;
        } else
        {
          temp_str += tmp.prefix;
          temp_str += "*";
          temp_str += m_image_file_format_info.delim;
        }
      } // if included
    } // for number of elements in filename
    
    temp_str.SetAt(temp_str.GetLength()-1,'.');
    temp_str += m_image_file_format_info.filetype;
    
    temp_date.month = -1;
    temp_date.day = -1;
    temp_date.year = -1;
    if(ff.FindFile(dir + temp_str))
    {
      ff.FindNextFile();
      temp_str = ff.GetFileName();      
      ii = GetImageInfo(temp_str);
      temp_date = ii.date;
      
    }
    dates_from_file.push_back(temp_date);
  }
  return dates_from_file;
}

ImageInfo ImageFileFormat::GetImageInfo(const CString& filename) const
{
  
  OrderElement tmpOE;            // temp order element
  CString expt, tube, initials;  // experiment name, tube id, initials
  int window = -1, session = -1;           // window #, session #
  int count = 0, year_size;
  Date date;                     // date
  
  CString tmpStr;                // tmp string
  CString fn = filename;         // non-const version of filename
  int plen, delim, elen;         // prefix length, location of delim, element length
  Date_Format d_format;          // temp varible used for date format
  
  d_format = m_image_file_format_info.m_date_format;
  
  // Remove path to file if it exists
  int tmpI1 = fn.ReverseFind('\\');
  int tmpI2 = fn.ReverseFind('/');
  
  // If path information exists
  if(!(tmpI1 == -1 && tmpI2 == -1))
  {
    // Set tmpI1 to location of last 'slash'
    if(tmpI1 < tmpI2) tmpI1 = tmpI2;
    
    // remove path information
    fn = filename.Right(filename.GetLength()-(tmpI1+1));
  }
  
  try {
    for(int i = 0; i < NUM_OF_ELEMENTS; i++)
    {
      tmpOE = m_image_file_format_info.InfoOrder[i];
      if(tmpOE.included)
      {
        // find next delimiter
        delim = fn.Find(m_image_file_format_info.delim);
        
        // set cut to end if last element
        if(delim == -1)
        { 
          if(i == m_image_file_format_info.numElementsInFileName-1) delim = fn.GetLength();
          else
          {
            TRACE("Error in GetImageInfo - ran out of elements too soon.");
          }
        }
        
        // get length of prefix
        plen = tmpOE.prefix.GetLength();
        
        // extract element minus prefix, set element length
        elen = delim-plen;
        tmpStr = fn.Mid(plen,elen);
        
        // set appropriate element
        int i;
        switch(tmpOE.element)
        {
        case EXPT: expt = tmpStr; break;
        case TUBE: tube = tmpStr; break;
        case WINDOW: window = atoi(tmpStr); break;
        case SESSION: session = atoi(tmpStr); break;
        case SDATE:
          
          // gets the date from file in the specified order
          for(i = 0; i < d_format.order.size(); i++)
          {
            if(d_format.order.at(i) == MONTH)
            {
              date.month = atoi(tmpStr.Mid(count,2));
              count += 3;                             // includes separator
            } else if(d_format.order.at(i) == DAY)
            {
              date.day = atoi(tmpStr.Mid(count,2));
              count += 3;                             // includes separator
            } else if(d_format.order.at(i) == YEAR)
            {
              
              if(d_format.four_year) year_size = 4;
              else year_size = 2;
              date.year = atoi(tmpStr.Mid(count,year_size));
              count += year_size + 1;                             // includes separator
            }
          }
          break;
        case INITIALS: initials = tmpStr; break;
        }
        
        // cut off previous element
        fn = fn.Right(fn.GetLength()-(delim+1));
        
      } // if element is included
    } // for number of elements in file
  }// try this
  catch (...) {
    TRACE("Error reading image file.\n\nThe file that caused the error was:\n'"+filename+"'");
  }
  
  // check to make sure that the required fields were read
  if (tube.GetLength() == 0 || window < 0 || session < 0)
  {
    static int message_ever_displayed = false;  // don't display message more than once (this is a bad way to do this, but it works)!!!
    
    if (!message_ever_displayed)
    {
      CString str;
      str.Format("Warning:  Image file '%s' is missing one or more of the necessary fields (tube, window, session)\r\n"
        "Unexpected results may occur.  It is recommended that you rename your image files.", 
        (const char*) filename);
      AfxMessageBox(str, MB_ICONWARNING | MB_OK);
      message_ever_displayed = true;
    }
    
    if (tube.GetLength() == 0)  tube = "001";
    if (window < 0)  window = 1;
    if (session < 0)  session = 1;
  }
  
  // Create ImageInfo instance and return
  ImageInfo ii;
  ii.expt = expt;
  ii.tube = tube;
  ii.window = window;
  ii.session = session;
  ii.date = date;
  ii.initials = initials;
  
  return ii;
}

ImageInfo ImageFileFormat::GetImageInfoOfFirstFile(const CString& directory) const
{
  CFileFind ff;
  if(ff.FindFile(directory + GetFileSearchString()))
  {
    ff.FindNextFile();
    return GetImageInfo(directory + ff.GetFileName());
  }
  else return ImageInfo();  // if file not found, return empty struct
}

CString ImageFileFormat::GetFirstImage(const CString& directory) const
{
  CFileFind ff;
  if(ff.FindFile(directory + GetFileSearchString()))
  {
    ff.FindNextFile();
    return directory + ff.GetFileName();
  }
  else return "";
}

// !!! This function does not seem to belong in this class -- STB 4/23/10
CRect ImageFileFormat::GetImageRectangle(const CString& filename)
{
  CRect rect(0,0,0,0);
  
  blepo::ImgBgr img;
  
  if(blepo::Load(filename,&img))
  {
    rect.right = img.Width();
    rect.bottom = img.Height();
  }
  assert(rect.right < 2000);
  return rect;
}

void ImageFileFormat::SetImageFileFormatInfo(const ImageFileFormatInfo& iffinfo)
{
  m_image_file_format_info = iffinfo;
}

const ImageFileFormatInfo& ImageFileFormat::GetImageFileFormatInfo() const
{
  return m_image_file_format_info;
}

bool ImageFileFormat::DateIsIncluded() const
{
  bool date_is_included = false;
  
  // must look through elements 1 by 1 until date element is found
  for(int i = 0; i < NUM_OF_ELEMENTS; i++)
  {
    if(m_image_file_format_info.InfoOrder[i].element == SDATE)
    { date_is_included = m_image_file_format_info.InfoOrder[i].included; break; }
  }
  
  return date_is_included;
}

void ImageFileFormat::SetToBartz()
{
  InitializeDateFormat();
  m_image_file_format_info.delim = '_';
  m_image_file_format_info.filetype = "jpg";
  m_image_file_format_info.numElementsInFileName = 7;
  
  vector<OrderElement> v;
  OrderElement x;
  
  x.element = EXPT;
  x.prefix = "";
  x.length = 0;
  x.included = TRUE;
  v.push_back(x);
  
  x.element = TUBE;
  x.prefix = "T";
  x.length = 3;
  v.push_back(x);
  
  x.element = WINDOW;
  x.prefix = "L";
  v.push_back(x);
  
  x.element = SDATE;
  x.prefix = "";
  x.length = 0;
  v.push_back(x);
  
  x.element = TIME;
  v.push_back(x);
  
  x.element = SESSION;
  x.length = 3;
  v.push_back(x);
  
  x.element = INITIALS;
  x.length = 0;
  v.push_back(x);
  
  x.element = MISC1;
  x.included = FALSE;
  v.push_back(x);
  
  x.element = MISC2;
  v.push_back(x);
  
  x.element = MISC3;
  v.push_back(x);
  
  x.element = MISC4;
  v.push_back(x);
  
  m_image_file_format_info.InfoOrder = v;
}

void ImageFileFormat::InitializeDateFormat()
{  
  // initialize date format       yyyy.mm.dd
  m_image_file_format_info.m_date_format.date_delim = ".";
  m_image_file_format_info.m_date_format.four_year = true;
  m_image_file_format_info.m_date_format.order.clear();
  m_image_file_format_info.m_date_format.order.push_back(YEAR);  
  m_image_file_format_info.m_date_format.order.push_back(MONTH);
  m_image_file_format_info.m_date_format.order.push_back(DAY);
  
}

void ImageFileFormat::Save(FILE* fp) const
{
  fprintf(fp,"%d %s %c",m_image_file_format_info.numElementsInFileName,
    (const char*) m_image_file_format_info.filetype,
    m_image_file_format_info.delim);
  
  OrderElement oe;
  for(int i = 0; i < NUM_OF_ELEMENTS; i++)
  {
    oe = m_image_file_format_info.InfoOrder[i];
    fprintf(fp," %d %s %d %d",oe.element,oe.prefix,oe.length,oe.included);
  }
  
  fprintf(fp,"\n");
  
  // saves date format
  // order-{elem1 elem2 elem3} four_year separator
  Date_Format d_format = m_image_file_format_info.m_date_format;
  fprintf(fp,"%d %d %d %d %s\n", d_format.order.at(0), d_format.order.at(1), d_format.order.at(2),
    d_format.four_year, d_format.date_delim);
}


// !!! EOF condition should only happen if our file save breaks mid way through or someone gives us a
// file with the wrong _Rootfly stamp.  This error should be caught earlier.
// The last condition this can occur is if someone randomly has a file with a _Rootfly stamp with a valid filesave
// namem but the file is not actually a Rootfly file.  Therefore, this check should be kept for this 'one in a 
// billion' chance -- ADP
bool ImageFileFormat::Load(FILE* fp)
{
  g_debug_logfile.Print("ImageFileFormat::Load() 1\n");
  if(!fp)return false;
  char buffer[1000];
  int i;
  ImageFileFormatInfo iffi;
  OrderElement oe;
  
  // store numElementsInFileName, filetype, and delimiter
  fscanf(fp, "%d", &iffi.numElementsInFileName);
  g_debug_logfile.Print("ImageFileFormat::Load() 2 neifn=%d\n", iffi.numElementsInFileName);
  fscanf(fp, "%1000s", buffer);
  iffi.filetype = buffer;
  g_debug_logfile.Print("ImageFileFormat::Load() 3 filetype=%s\n", iffi.filetype);
  fscanf(fp, "%1000s", buffer);
  iffi.delim = buffer[0];
  g_debug_logfile.Print("ImageFileFormat::Load() 4 delim=%c\n", iffi.delim);
  
  // store name, element, prefix, length, and included for each Element(expt, tube, etc)
  // fgetc bc the prefix may be the empty string ""
  for(i = 0; i < NUM_OF_ELEMENTS; i++)
  {
    fscanf(fp, "%d", &oe.element);
    g_debug_logfile.Print("ImageFileFormat::Load() 5 elem=%d\n", oe.element);
    char c = fgetc(fp);       // skip the first space
    oe.prefix = "";

    for(;;)
    {
      c = fgetc(fp);
      if (c == EOF)  return false;  // ERROR:  end of file reached
      if (c == ' ')  break;
      oe.prefix += c;
    }

    // !!! Warning: scanning a bool directly seems to work, but is it guaranteed?
    fscanf(fp, "%d %d", &oe.length, &oe.included);
    iffi.InfoOrder.push_back(oe);
    g_debug_logfile.Print("ImageFileFormat::Load() 6 len=%d inc=%d\n", oe.length, oe.included);
  }
  
  m_image_file_format_info = iffi;
  
  // loads the date format
  int temp1,temp2,temp3;
  bool temp4;
  Date_Format d_format;
//  char temp5[1000];  // old (replaced with buffer) -- STB 4/27/10
  // !!! Warning: scanning a bool directly seems to work, but is it guaranteed?
  fscanf(fp,"%d %d %d %d %1000s\n", &temp1, &temp2, &temp3, &temp4, &buffer);
  g_debug_logfile.Print("ImageFileFormat::Load() 7 date1=%d date2=%d date3=%d date4=%d dataformat=%s\n", temp1, temp2, temp3, temp4, buffer);
  d_format.date_delim = buffer;
  d_format.order.push_back(temp1);
  d_format.order.push_back(temp2);
  d_format.order.push_back(temp3);
  d_format.four_year = temp4;
  m_image_file_format_info.m_date_format = d_format;
  
  return true;
}
