/*This class will handle all of the images and thumbnails*/

#include "DataStructures.h"

// uncomment this to use the cache.  The cache is not recommended at the moment, because it just slows things down.
// I realize this is ironic, but the cache forces us to copy the image after loading it.  There may be other 
// inefficiencies as well.
// We need to reimplement this so that it just stores the most recently accessed images and avoids unnecessary
// image copying. -- STB 5/3/10
// #define USE_CACHE 

// Constructor and Destructor
ImageManager::ImageManager() : m_image_directory(""), m_current_tube(""), m_current_window(-1), m_current_session(-1)
{}
ImageManager::~ImageManager()
{}

void ImageManager::Initialize( const CString& dir, const ImageFileFormat& iff, const SessionCollapser& sess_coll )
{
  m_image_directory = dir; m_image_file_format = iff; m_session_collapser = sess_coll;

#ifdef USE_CACHE 
  if ( dir != m_image_directory )
  {
    LoadAll( m_current_tube, m_current_window, m_current_session );
  }
#endif // USE_CACHE
}

// Stores common images and managers cache size
void ImageManager::SetCurrentTWS(const CString& tube, int window, int session)
{

  // Make sure an image directory and file format have been specified
  if ( m_image_directory == "" )
  {
    return;
  }

  // Check input TWS against stored TWS to see which images need to be loaded from disk
  // This assumes that only either W or S changes at a time.  If T changes, the whole thing needs to be reloaded anyway.
  // If both W and S change at once, this will fail.

  // If TWS are the same as before, no change to the cache is needed
  if ( m_current_tube == tube && m_current_window == window && m_current_session == session )
  {
    return;
  }
  
  if ( window == -1 || session == -1 )
  {
    return;
  }

#ifdef USE_CACHE 
  // Session has shifted less than NUM_THUMBNAILS
  if ( tube == m_current_tube && abs( session - m_current_session ) < NUM_THUMBNAILS && session != m_current_session )
  {
    int s_offset = session - m_current_session;
    CString filename = "";

    for ( int i = -NUM_THUMBNAILS/2; i <= -1; i++ )
    {
      for ( int j = -NUM_THUMBNAILS/2; j <= NUM_THUMBNAILS/2; j++ )
      {
        if ( InCache( i,j ) )
        {
          filename = m_image_file_format.GetImageFileName( m_image_directory, tube, window + i, session + m_session_collapser.GetRemappedIndex( j ) );
          blepo::Load( filename, &m_image_cache[ j + NUM_THUMBNAILS/2 ][ i + NUM_THUMBNAILS/2 ] );
        }
      }
    }
    if ( s_offset > 0 )
    {
      for ( int j = -NUM_THUMBNAILS/2; j < NUM_THUMBNAILS/2 - s_offset; j++ )
      {
        m_image_cache[ j + NUM_THUMBNAILS/2 ][ NUM_THUMBNAILS/2 ] = m_image_cache[ j + NUM_THUMBNAILS/2 + s_offset ][ NUM_THUMBNAILS/2 ];
      }
      for ( j = NUM_THUMBNAILS/2 - s_offset; j <= NUM_THUMBNAILS/2; j++ )
      {
        filename = m_image_file_format.GetImageFileName( m_image_directory, tube, window, session + m_session_collapser.GetRemappedIndex( j ) );
        blepo::Load( filename, &m_image_cache[ j + NUM_THUMBNAILS/2 ][ NUM_THUMBNAILS/2 ] );
      }
    }
    else
    {
      for ( int j = NUM_THUMBNAILS/2; j > -NUM_THUMBNAILS/2 - s_offset; j-- )
      {
        m_image_cache[ j + NUM_THUMBNAILS/2 ][ NUM_THUMBNAILS/2 ] = m_image_cache[ j + NUM_THUMBNAILS/2 + s_offset ][ NUM_THUMBNAILS/2 ];
      }
      for ( j = -NUM_THUMBNAILS/2 - s_offset; j >= -NUM_THUMBNAILS/2; j-- )
      {
        filename = m_image_file_format.GetImageFileName( m_image_directory, tube, window, session + m_session_collapser.GetRemappedIndex( j ) );
        blepo::Load( filename, &m_image_cache[ j + NUM_THUMBNAILS/2 ][ NUM_THUMBNAILS/2 ] );
      }
    }
    for ( i = 1; i <= NUM_THUMBNAILS/2; i++ )
    {
      for ( int j = -NUM_THUMBNAILS/2; j <= NUM_THUMBNAILS/2; j++ )
      {
        if ( InCache( i,j ) )
        {
          filename = m_image_file_format.GetImageFileName( m_image_directory, tube, window + i, session + m_session_collapser.GetRemappedIndex( j ) );
          blepo::Load( filename, &m_image_cache[ j + NUM_THUMBNAILS/2 ][ i + NUM_THUMBNAILS/2 ] );
        }
      }
    }
  }

  // Window has shifted less than NUM_THUMBNAILS
  else if ( tube == m_current_tube && abs( window - m_current_window ) < NUM_THUMBNAILS )
  {
    int w_offset = window - m_current_window;
    CString filename = "";

    for ( int i = -NUM_THUMBNAILS/2; i <= -1; i++ )
    {
      for ( int j = -NUM_THUMBNAILS/2; j <= NUM_THUMBNAILS/2; j++ )
      {
        if ( InCache( j,i ) )
        {
          filename = m_image_file_format.GetImageFileName( m_image_directory, tube, window + j, session + m_session_collapser.GetRemappedIndex( i ) );
          blepo::Load( filename, &m_image_cache[ i + NUM_THUMBNAILS/2 ][ j + NUM_THUMBNAILS/2 ] );
        }
      }
    }
    if ( w_offset > 0 )
    {
      for ( int j = -NUM_THUMBNAILS/2; j < NUM_THUMBNAILS/2 - w_offset; j++ )
      {
        m_image_cache[ NUM_THUMBNAILS/2 ][ j + NUM_THUMBNAILS/2 ] = m_image_cache[ NUM_THUMBNAILS/2 ][ j + NUM_THUMBNAILS/2 + w_offset ];
      }
      for ( j = NUM_THUMBNAILS/2 - w_offset; j <= NUM_THUMBNAILS/2; j++ )
      {
        filename = m_image_file_format.GetImageFileName( m_image_directory, tube, window + j, session + m_session_collapser.GetRemappedIndex( 0 ) );
        blepo::Load( filename, &m_image_cache[ NUM_THUMBNAILS/2 ][ j + NUM_THUMBNAILS/2 ] );
      }
    }
    else
    {
      for ( int j = NUM_THUMBNAILS/2; j > -NUM_THUMBNAILS/2 - w_offset; j-- )
      {
        m_image_cache[ NUM_THUMBNAILS/2 ][ j + NUM_THUMBNAILS/2 ] = m_image_cache[ NUM_THUMBNAILS/2 ][ j + NUM_THUMBNAILS/2 + w_offset ];
      }
      for ( j = -NUM_THUMBNAILS/2 - w_offset; j >= -NUM_THUMBNAILS/2; j-- )
      {
        filename = m_image_file_format.GetImageFileName( m_image_directory, tube, window + j, session + m_session_collapser.GetRemappedIndex( 0 ) );
        blepo::Load( filename, &m_image_cache[ NUM_THUMBNAILS/2 ][ j + NUM_THUMBNAILS/2 ] );
      }
    }
    for ( i = 1; i <= NUM_THUMBNAILS/2; i++ )
    {
      for ( int j = -NUM_THUMBNAILS/2; j <= NUM_THUMBNAILS/2; j++ )
      {
        if ( InCache( j,i ) )
        {
          filename = m_image_file_format.GetImageFileName( m_image_directory, tube, window + j, session + m_session_collapser.GetRemappedIndex( i ) );
          blepo::Load( filename, &m_image_cache[ i + NUM_THUMBNAILS/2 ][ j + NUM_THUMBNAILS/2 ] );
        }
      }
    }
  }
  
  else 
  {
    LoadAll( tube, window, session );  
  }
#endif // USE_CACHE

  m_current_tube = tube;
  m_current_window = window;
  m_current_session = session;
  
}

// Gets image from either cache or disk
// Returns false is there is no image matching the specified tube, window, session
// Returns true otherwise
bool ImageManager::GetImage(const CString& tube, int window, int session, blepo::ImgBgr *out) const
{

  if ( out == NULL )
  {
    return false;
  }

  // Make sure an image directory and file format have been specified
  if ( m_image_directory == "" )
  {
    return false;
  }

  // Make sure image exists
  // If not, return false and load no image
  CString filename;
  filename = m_image_file_format.GetImageFileName( m_image_directory, tube, window, session );
  if ( filename == "" )
  {
    return false;
  }

#ifdef USE_CACHE 
  // If image is in cache, return it from there
  if ( tube == m_current_tube && InCache( window - m_current_window, session - m_current_session ) )
  {
    *out = m_image_cache[ session - m_current_session + NUM_THUMBNAILS/2 ][ window - m_current_window + NUM_THUMBNAILS/2 ];
    return true;
  }

  // Otherwise, load from disk
  else
  {
    return blepo::Load( filename, out );
  }

#else // USE_CACHE

  return blepo::Load( filename, out );

#endif // USE_CACHE
}

// Should return smaller downsampled version of image
// Currently, same as GetImage
bool ImageManager::GetThumbnail(const CString& tube, int window, int session, blepo::ImgBgr* out) const
{
  return GetImage( tube, window, session, out );
}

// Checks whether or not an image exists
// Returns true if so, false otherwise
bool ImageManager::ImageExists(const CString& tube, int window, int session) const
{
  CString filename;
  filename = m_image_file_format.GetImageFileName( m_image_directory, tube, window, session );
  if ( filename != "" )
  {
    return true;
  }
  else
  {
    return false;
  }
}

bool ImageManager::InCache( const int w_offset, const int s_offset ) const
{
  if (  ( w_offset == 0 && abs( s_offset ) <= NUM_THUMBNAILS/2 ) ||
        ( s_offset == 0 && abs( w_offset ) <= NUM_THUMBNAILS/2 ) ||
        ( abs( w_offset ) <= GRID_SIZE/2 && abs( s_offset ) <= GRID_SIZE/2 ) )
  {
    return true;
  }
  else
  {
    return false;
  }
}


void ImageManager::LoadAll( const CString& tube, const int window, const int session )
{
#ifdef USE_CACHE
  CString filename = "";
  for ( int i = -NUM_THUMBNAILS/2; i <= NUM_THUMBNAILS/2; i++ )
  {
    for ( int j = -NUM_THUMBNAILS/2; j <= NUM_THUMBNAILS/2; j++ )
    {
      if ( InCache( j,i ) )
      {
        filename = m_image_file_format.GetImageFileName( m_image_directory, tube, window + j, session + m_session_collapser.GetRemappedIndex( i ) );
        blepo::Load( filename, &m_image_cache[ i + NUM_THUMBNAILS/2 ][ j + NUM_THUMBNAILS/2 ] );
      }
    }
  }
#endif // USE_CACHE
}